/*
 * decode.c
 *
 * This is really where all the fun happens...
 *
 */

#include "sim.h"

/*
 * 
 * Return the new PC if we commited a jump, else return 0 and the sequencer
 * will autoincrement for us.
 *
 */
unsigned int decode()
{
  unsigned int opcode;

  opcode = curdsp.sequencer.curinstr;
  /* No-oper codes... */
  if ((opcode>>16) == 0x0000)
    return 0; /* NOP */
  if ((opcode>>4) == 0x02800) /* idle */
    {
      if (opcode & 0xf)
	{
	  /*
	   * idle(n) waits n cycles or for an interrupt, whichever is first 
	   *
	   * FIXME: above theory is wrong.  Hey mid, read the manuals before
	   * making statements like that.
	   */
	  int n = opcode & 0xf;
	  while (n  && (irq_checkdone()<0))
	    {
	      idler();
	      n--;
	    }
	  return 0;
	}
      else
	{
	  /*
	   * idle waits indefinitly for an interrupt
	   */
	  while (irq_checkdone()<0)
	    {
	      idler(); 
	    }
	  return 0;
	}
      return 0;
    }

  /* full codes... */
  if (opcode & BIT23) /* 1 */
    {
      if (opcode & BIT22) /* 11 */
	{
	  if (opcode & BIT21) /* 111 */
	    {
	      fprintf(stderr, "decode: unknown: 111\n");
	    }
	  else /* 110 */
	    {
	      fprintf(stderr, "decode: unknown: 110\n");
	    }
	}
      else /* 10 */
	{
	  if (opcode & BIT21) /* 101 */
	    {
	      if (!setdmindirect((opcode>>20)&0x1, 
				 (opcode>>2)&0x3, 
				 opcode&0x3, 
				 (opcode>>4)&0x1ffff))
		fprintf(stderr,"decode: indirect data memory set failed\n");
	    }
	  else /* 100 */
	    {
	      if (opcode & BIT20) /* 1001 */
		{
		  /* dm(addr) = reg */
		  setdmfromreg((opcode >> 4) & 0x3fff,   /* address */
			       getregptr_rgp(opcode & 0xf, /* reg */
					     (opcode >> 18) & 0x3)); /* rgp */
		}
	      else /* BIT20 off */
		{
		  /* reg = dm(addr) */
		  setregfromdm((opcode>>4)&0x3fff, /* addr */
			       getregptr_rgp(opcode & 0xf, /* reg */
					     (opcode>>18)&0x3)); /* rgp */
		  
		  /* set the internal _B register if its an I reg */
		  if ( ((opcode>>18&0x3)>0) && ((opcode>>18)&0x3)<3)
		    {
		      if ((opcode&0xf)<4)
			setdagregrgp((opcode&0xf), ((opcode>>18)&0x3)+3, (opcode>>4)&0x3fff);
		    }
		}
	    }
	}
    }
  else /* BIT23 off */
    {
      if (opcode & BIT22) /* 01 */
	{
	  if (opcode & BIT21) /* 011 */
	    {
	      /* dm(i, m) = dreg */
	      unsigned int dag = (opcode & BIT20)?1:0;
	      if ((opcode & BIT19) && !(opcode & BIT18))
		{
		  unsigned int amf = (opcode >> 13) & 0x1f;
                  unsigned int i = (opcode >> 2) & 0x3;
                  unsigned int m = opcode & 0x3;
		  unsigned int dreg = (opcode >> 4) & 0xf;
		  setdmfromreg_indirect(dag, i, m, getregptr_rgp(dreg,0));
		  if (amf) fprintf(stderr, "decode: AMF unhandled\n");
#if 0
		  fprintf(stderr, "decode: unknown: dm(i%d, m%d) = %d  %s\n", 
			  dag?i+4:i, 
			  dag?m+4:m, 
			  /*regs[dreg][0], */ dreg,
			  amf?"(AMF unhandled)":"");
#endif
		}
	      else /* no amf */
		{
		  unsigned int data = (opcode >> 4) & 0xff;
		  unsigned int i = (opcode >> 1) & 0x1;
		  unsigned int m = opcode & 0x1;
		  fprintf(stderr, "decode: unknown: dm(i%d, m%d) = 0x%02x\n", i, m, data); 
		}
	    }
	  else /* 010 */
	    {
	      if (opcode & BIT20) /* 0101 */
		{
		  if (opcode & BIT19) /* 0101 1 */
		    {
		      if (opcode & BIT18) /* 0101 11 */
			{
			  fprintf(stderr, "decode: unknown: 0101 11\n");
			}
		      else /* 0101 10 */
			{
#if 0
			  unsigned int amf = (opcode >> 13) & 0x1f;
			  unsigned int dreg = (opcode >> 4) & 0xf;
			  unsigned int i = (opcode >> 1) & 0x1;
			  unsigned int m = opcode & 0x1;
#endif
			  fprintf(stderr, "decode: unknown: 0101 10\n");
 			}
		    }
		  else /* 01010 */
		    {
		      fprintf(stderr, "decode: unknown: 0101 0\n");
		    }
		}
	      else /* 0100 */
		setregrgp(opcode & 0xf, 0, (opcode>>4)&0x3fff);
	    }
	}
      else  /* 00 */
	{
	  if (opcode & BIT21) /* 001 */
	    {
	      if (opcode & BIT20) /* 0011 */
		{
		  setregrgp(opcode & 0xf, /* reg */
			    (opcode>>18) & 0x3, /* rgp */
			    (opcode>>4) & 0x3fff /* addr */);

		  /* set the internal _B register if its an I reg */
		  if ( ((opcode>>18&0x3)>0) && ((opcode>>18)&0x3)<3)
		    {
		      if ((opcode&0xf)<4)
			setdagregrgp((opcode&0xf), ((opcode>>18)&0x3)+3, (opcode>>4)&0x3fff);
		    }
		}
	      else /* 0010 */
		{
		  fprintf(stderr, "decode: unknown: 0010\n");
		}
	    }
	  else /* 000 */
	    {
	      if (opcode & BIT20) /* 0001 */
		{
		  if (opcode & BIT19) /* 00011 */
		    {
		      if (opcode & BIT18) /* 000111 */
			{			  
			  if (testcond(opcode&0xf)) /* call */
			    {
			      pc_push();
			      return (opcode >> 4) & 0x3fff;
			    }
			}
		      else /* 000110 */
			{
			  if (testcond(opcode&0xf)) /* jump */
			    return (opcode >> 4) & 0x3fff;
			}
		    }
		  else /* 00010 */
		    {
		      if (opcode & BIT18) /* 000101 */
			{
			  unsigned short addr,cond;
			  addr = (opcode>>4) & 0x3fff;
			  cond = opcode & 0xf;
			  curdsp.sequencer.pc++;
			  pc_push();
			  loop_push(addr, cond);
			  curdsp.sequencer.pc--;
			}
		      else /* 000100 */
			fprintf(stderr, "decode: unknown: 000100\n");
		    }
		}
	      else /* 0000 */
		{
		  if (opcode & BIT19) /* 0000 1 */
		    {
		      if (opcode & BIT18) /* 0000 11 */
			{
			  if (opcode & BIT17) /* 0000 111 */
			    {
			      if (opcode & BIT16) /* 0000 1111 */
				{
#if 0
				  unsigned int sf = (opcode >> 11) & 0xf;
				  unsigned int xop = (opcode >> 8) & 0x7;
				  unsigned int exp = opcode & 0xff;
#endif
				  fprintf(stderr, "decode: unknown: 0000 1111\n");
#if 0
				  sprintf(buf, "sr = %s %s %s by 0x%04x %s", 
					  sfs[sf][2][0]==1?"sr or":"",
					  sfs[sf][0],
					  xops[xop],
					  exp,
					  sfs[sf][1]);
#endif
				}
			      else /* 0000 1110 */
				{
				  fprintf(stderr, "decode: unknown: 0000 1110\n");
				}
			    }
			  else /* 0000 110 */
			    {
			      if (opcode & BIT16)
				regmove((opcode>>4)&0xf, (opcode>>10)&0x3, opcode&0xf, (opcode>>8)&0x3);
			      else /* 0000 1100 */
				{
				  if (((opcode>>2)&03)>=2)
				    {
				      if (((opcode>>2)&0x3)==2)
					set_mstatclearbit(MSTAT_GMODE);
				      else
					set_mstatbit(MSTAT_GMODE);
				    }
				  if (((opcode>>4)&03)>=2)
				    {
				      if (((opcode>>4)&0x3)==2)
					set_mstatclearbit(MSTAT_SECREG);
				      else
					set_mstatbit(MSTAT_SECREG);
				    }

				  if (((opcode>>6)&03)>=2)
				    {
				      if (((opcode>>6)&0x3)==2)
					set_mstatclearbit(MSTAT_BITREV);
				      else
					set_mstatbit(MSTAT_BITREV);
				    }

				  if (((opcode>>8)&03)>=2)
				    {
				      if (((opcode>>8)&0x3)==2)
					set_mstatclearbit(MSTAT_AVLATCH);
				      else
					set_mstatbit(MSTAT_AVLATCH);
				    }

				  if (((opcode>>0x0a)&03)>=2)
				    {
				      if (((opcode>>0x0a)&0x3)==2)
					set_mstatclearbit(MSTAT_ARSAT);
				      else
					set_mstatbit(MSTAT_ARSAT);
				    }

				  if (((opcode>>0x0c)&03)>=2)
				    {
				      if (((opcode>>0x0c)&0x3)==2)
					set_mstatclearbit(MSTAT_MMODE);
				      else
					set_mstatbit(MSTAT_MMODE);
				    }

				  if (((opcode>>0x0e)&03)>=2)
				    {
				      if (((opcode>>0x0e)&0x3)==2)
					set_mstatclearbit(MSTAT_TIMER);
				      else
					set_mstatbit(MSTAT_TIMER);
				    }
				}
				
			    }
			}
		      else /* 0000 10 */
			{
			  if ((opcode >> 4) == 0xa001) /* rti */
			    {
			      if (testcond(opcode&0xf))
				seq_rti();
			    }
			  else if ((opcode >> 4) == 0xa000) /* rts */
			    {
			      if (testcond(opcode&0xf))
				{	      
				  pc_pop();
				  return ++curdsp.sequencer.pc;
				}
			    }
			  else if ((opcode >> 4) == 0x09080)
			    {
#if 0
			      unsigned int dag = (opcode >> 4) & 0x1;
			      unsigned int i = (opcode >> 2) & 0x3;
			      unsigned int m = opcode & 0x3;
#endif
			      fprintf(stderr, "decode: modify: not implemented\n");
#if 0
			      sprintf(buf, "modify (i%d, m%d)", dag?i+4:i, dag?m+4:m);
#endif
			    }
			  else if ((opcode>>8) == 0xb00)
			    {
			      unsigned int ireg = (opcode>>6)&0x3;
			      unsigned int dir = (opcode>>4)&0x1;
			      unsigned int cond = (opcode & 0xf);
			      
			    }
			}
		    }
		  else /* 0000 0 */
		    {
		      if (opcode & BIT18) /* 0000 01 */
			{
			  if ((opcode >> 5) == 0x2000)
			    {
			      /* stack control */
			      fprintf(stderr, "decode: unknown: [stack control]");
			    }

			  if (opcode & BIT17) /* 0000 011 */
			    {
			      if (opcode & BIT16) /* 0000 0111 */
				{
				  /* DIVQ */
#if 0
				  unsigned int xop = (opcode >> 8) & 0x7;
#endif
				  fprintf(stderr, "decode: divq: not implemented\n");
#if 0
				  sprintf(buf, "divq %s", xops[xop]);
#endif
				}
			      else /* 0000 0110 */
				{
				  /* DIVS */
#if 0
				  unsigned int yop = (opcode >> 11) & 0x3;
				  unsigned int xop = (opcode >> 8) & 0x7;
#endif
				  fprintf(stderr, "decode: divs: not implemented\n");
#if 0
				  sprintf(buf, "divs %s, %s", yops[yop], xops[xop]);
#endif
				}
			    }
			  else /* 0000 010 */
			    {
			      fprintf(stderr, "decode: unknown: 0000 010\n");
			    }
			}
		      else /* 000000 */
			{
			  if (opcode & BIT17) /* 0000 001 */
			    {
			      if (opcode & BIT16) /* 0000 0011 */
				{
				  /* call/jump on flag in */
#if 0
				  unsigned int addr = (opcode >> 4) & 0xfff;
				  unsigned int addrmsb = (opcode >> 2) & 0x3;
				  unsigned int fic = (opcode >> 1) & 0x1;
				  unsigned int s = (opcode & 0x1);
#endif
				  fprintf(stderr, "decode: conditional jump on flag in: not implemented\n");
#if 0
				  sprintf(buf, "if %sfi %s 0x%04x",
					  fic?"":"not ",
					  s?"call":"jump",
					  (addrmsb<<12)+addr);
#endif
		     
				}
			      else /* 0000 0010 */
				{
				  /* modify flag out */
#if 0
				  unsigned int fl[4] = {
				    (opcode >> 10) & 0x3,
				    (opcode >> 8) & 0x3,
				    (opcode >> 6) & 0x3,
				    (opcode >> 4) & 0x3,
				  };
				  unsigned int cond = opcode & 0xf;
				  int cur = 0, i = 0;
#endif
				  fprintf(stderr, "decode: mode: not implemented\n");
#if 0
				  if (cond < 0xf)
				    {
				      sprintf(buf, "if %s ", conds[cond]);
				      i += 6;
				    }
				  for (cur=0;cur<3;cur++)
				    {
				      if (fl[cur])
					{
					  sprintf(&(buf[i]), "%s fl%d ",
						  flagops[fl[cur]],
						  cur);
					  i += strlen(flagops[fl[cur]]) + 5;
					}
				    }
				  if (fl[3])
				    {
				      sprintf(&(buf[i]), "%s flag_out", 
					      flagops[fl[3]]);
				    }
				 
#endif 
				}
			    }
			  else /* 0000 000 */
			    {
			      if (opcode & BIT16) /* 0000 0001 */
				{
#if 0
				  unsigned int addr = (opcode >> 4) & 0x7ff;
				  unsigned int dreg = opcode & 0xf;
#endif
				  unsigned int dir = opcode & BIT15;
				 
				  if (dir)
				    fprintf(stderr, "decode: io()=reg: not implemented\n");
				  else
				    fprintf(stderr, "decode: reg=io(): not implemented\n");
#if 0
				  if (dir)
				    sprintf(buf, "io(%d) = %s", addr, regs[dreg][0]);
				  else
				    sprintf(buf, "%s = io(%d)", regs[dreg][0], addr);
#endif
				}
			      else /* 0000 0000 */
				{
				  fprintf(stderr, "decode: unknown: 0000 0000\n");
				}
			    }
			}
		    }
		}
	    }
	}
    }

  return 0;
}
